import LFUCache from '@akashbabu/lfu-cache';
import { DLL } from '@akashbabu/node-dll';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */
function __awaiter(thisArg, _arguments, P, generator) {
  return new (P || (P = Promise))(function (resolve, reject) {
    function fulfilled(value) {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    }

    function rejected(value) {
      try {
        step(generator["throw"](value));
      } catch (e) {
        reject(e);
      }
    }

    function step(result) {
      result.done ? resolve(result.value) : new P(function (resolve) {
        resolve(result.value);
      }).then(fulfilled, rejected);
    }

    step((generator = generator.apply(thisArg, _arguments || [])).next());
  });
}

/* istanbul ignore file */
var LEVEL;
(function (LEVEL) {
    LEVEL[LEVEL["DEBUG"] = 0] = "DEBUG";
    LEVEL[LEVEL["INFO"] = 1] = "INFO";
    LEVEL[LEVEL["WARN"] = 2] = "WARN";
    LEVEL[LEVEL["ERROR"] = 3] = "ERROR";
})(LEVEL || (LEVEL = {}));
/**
 * Custom Logger with different log levels
 * and enable/disable logging
 */
class Logger {
    constructor(level = LEVEL.DEBUG, printMsg = false) {
        this.level = level;
        this.printMsg = printMsg;
        this.prefix = '';
    }
    /**
     * Prints debug messages to console
     * @param args Log msgs
     */
    debug(...args) {
        this.log(LEVEL.DEBUG, ...args);
    }
    /**
     * Prints info messages to console
     * @param args Log msgs
     */
    info(...args) {
        return this.log(LEVEL.INFO, ...args);
    }
    /**
     * Prints warn messages to console
     * @param args Log msgs
     */
    warn(...args) {
        return this.log(LEVEL.WARN, ...args);
    }
    /**
     * Prints error messages to console
     * @param args Log msgs
     */
    error(...args) {
        return this.log(LEVEL.ERROR, ...args);
    }
    /**
     * Enable writing logs to console
     */
    enable() {
        this.printMsg = true;
    }
    /**
     * Disable writing logs to console
     */
    disable() {
        this.printMsg = false;
    }
    /**
     * Adds this prefix for every logged line
     * This is especially useful when the library
     * name has to be printed on each logged line
     *
     * @param prefix Prefix for every log
     */
    setPrefix(prefix) {
        this.prefix = prefix;
    }
    log(logLevel, ...args) {
        let print = console.log;
        if (this.level === LEVEL.WARN) {
            print = console.warn;
        }
        else if (this.level === LEVEL.ERROR) {
            print = console.error;
        }
        if ((logLevel <= this.level) && this.printMsg) {
            print(this.prefix, ...args);
        }
        else if (logLevel >= LEVEL.WARN) {
            print(this.prefix, ...args);
        }
    }
}
Logger.LEVEL = LEVEL;

/**
 * This error will be thrown when an invalid wildcards sequence
 * is detected
 */
class InvalidPatternError extends Error {
    constructor() {
        super('DO NOT use **/*, */**, **/** in your event string because it is equivalent to using /**/');
        this.name = this.constructor.name;
    }
}

const SEPERATOR = '/';
/**
 * Predicate that checks if the given
 * part is a wildcard part or not
 *
 * @param part Part of the event to check
 *
 * @returns {Boolean}
 */
function isWildcard(part) {
    return part === '*' || part === '**';
}
/**
 * Returns the parts of the event.
 * i.e, it essentially splits the event string
 * based on a seperator '/' and removes
 * the empty parts
 *
 * @param event Event
 *
 * @returns list of event parts
 */
const getEventParts = (event) => sanitizeEvent(event, true);
/**
 * Predicate that checks if any part of the
 * event is a wildcard or not
 *
 * @param event Event to check for wildcard
 *
 * @returns true if the given event contains
 *    wildcards
 */
function isPatternEvent(event) {
    return !!getEventParts(event).find(isWildcard);
}
/**
 * Checks if the given event is valid
 *
 * @param event event to check
 *
 * @returns {Boolean}
 */
function isValidEvent(event) {
    const eventParts = getEventParts(event);
    // let prevPart = eventParts[0];
    let prevPart = '';
    return !eventParts.some((eventPart) => {
        if (isWildcard(prevPart) && isWildcard(eventPart)) {
            // This essentially checks for */** or **/* or **/**
            if ((eventPart + prevPart).length > 2)
                return true;
        }
        prevPart = eventPart;
    });
}
/**
 * Closure that returns the next patterPart in the array
 *
 */
function arrayIterator(arr) {
    let index = -1;
    return (i) => {
        if (i)
            index = i + 1;
        else
            index++;
        return [arr[index], index];
    };
}
/**
 * Checks if the pubEvent matches the pattern event
 *
 * @param pubEvent Event to check for match
 * @param subEvent pattern subEvent
 *
 * @returns true if it matches
 *
 * @throws Invalid Pattern if it violates wildcard
 *     policies
 */
function doesPatternMatch(pubEvent, subEvent) {
    if (!isValidEvent(subEvent))
        throw new InvalidPatternError();
    // Publish event parts
    const pubParts = getEventParts(pubEvent);
    const getNextPubPart = arrayIterator(pubParts);
    // Subscribe event parts
    const subParts = getEventParts(subEvent);
    const getNextSubPart = arrayIterator(subParts);
    /**
     * Checks if the parts matches post a '**'
     * if not then, checks for match from
     * the next eventPart index
     *
     * @param eventPart Event part in the form of
     *       [eventPart, eventIndex]
     * @param nextPatternIndex Next Pattern index
     *
     * @returns true if parts match after '**'
     */
    function checkMulti([eventPart, eventIndex], nextPatternIndex) {
        if (!eventPart)
            return false;
        const isMatch = match([eventPart, eventIndex], getNextSubPart(nextPatternIndex - 1));
        return !isMatch ? checkMulti(getNextPubPart(eventIndex), nextPatternIndex) : true;
    }
    function match([eventPart, eventIndex] = getNextPubPart(), [patternPart] = getNextSubPart()) {
        // if eventPart is present but not patternPart
        // or vice-versa then it's not a match
        if ((!eventPart && patternPart && patternPart !== '**') || (!patternPart && eventPart))
            return false;
        // if both parts are not present then
        // it's a match
        if (!eventPart && !patternPart)
            return true;
        // if it's not a pattern (*, **) then perform
        // a direct string match
        if (!isWildcard(patternPart)) {
            return patternPart === eventPart ? match() : false;
        }
        // Since * would match any string, goto next match;
        if (patternPart === '*')
            return match();
        // Else patternPart == '**'
        const [nextPatternPart, nextPatternIndex] = getNextSubPart();
        return nextPatternPart ? checkMulti([eventPart, eventIndex], nextPatternIndex) : true;
    }
    return match();
}
/**
 * Sanitizes the given event by removing all
 * empty parts. If `getParts` is true, then it
 * returns all the non-empty parts as a list else
 * string
 *
 * @param event Event
 * @param getParts If true, then returns non-empty parts list
 *
 * @returns sanitized-event
 */
function sanitizeEvent(event, getParts) {
    const sanitizedEvent = [];
    let part = '';
    for (const char of event) {
        if (char === SEPERATOR) {
            if (part) {
                sanitizedEvent.push(part);
                part = '';
            }
        }
        else {
            part += char;
        }
    }
    // Check if there are any remaining characters
    // in `part`, if so then add them up
    if (part) {
        sanitizedEvent.push(part);
    }
    return getParts ? sanitizedEvent : sanitizedEvent.join(SEPERATOR);
}

function mergeOptions(givenOpts = {}, defaultOpts = {}) {
    return [...new Set([
            ...Object.keys(givenOpts),
            ...Object.keys(defaultOpts),
        ])].reduce((result, key) => {
        if (typeof givenOpts[key] === 'object') {
            result[key] = mergeOptions(givenOpts[key], defaultOpts[key]);
        }
        else {
            result[key] = givenOpts[key] !== undefined
                ? givenOpts[key]
                : defaultOpts[key];
        }
        return result;
    }, {});
}

/**
 * @hidden
 *
 * Task Runner runs only the
 * controlled number of tasks at any given
 * instance of time.
 *
 * @param worker function to be called for executing the task
 * @param options
 *
 * @returns Task adder function
 */
class TaskQueue {
    constructor(worker, options = {}) {
        this.worker = worker;
        this.tasks = [];
        this.running = 0;
        this.options = mergeOptions(options, {
            maxRunners: 10,
        });
    }
    add(...args) {
        return new Promise((resolve) => {
            // add task to DLL
            this.tasks.push({
                args,
                resolve,
            });
            // inform runner about the new task
            this.runner();
        });
    }
    /**
     * This is responsible for running the tasks
     * in parallel and to maintain only the
     * limited number of concurrent tasks
     */
    runner() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.running < this.options.maxRunners) {
                this.running++;
                const task = this.tasks.shift();
                if (task) {
                    const { args, resolve } = task;
                    try {
                        const data = yield this.worker(...args);
                        resolve([null, data]);
                    }
                    catch (error) {
                        resolve([error]);
                    }
                    this.running--;
                    // assuming a new task might be present,
                    // we're calling the runner again.
                    this.runner();
                }
                else {
                    // since no task has been found,
                    // we'll exit the runner process
                    this.running--;
                }
            }
        });
    }
}
// export default function TaskQueue(worker: (...args: any[]) => any, { maxRunners = 10 }: IOptions = {}): IRunTask {
//   const tasks: ITask[] = [];
//   // number of currently running tasks
//   const running: number = 0;
//   runner();
//   return {
//     add(...args: any[]) {
//       return new Promise((resolve) => {
//         // add task to DLL
//         tasks.push({
//           args,
//           resolve,
//         });
//         // inform runner about the new task
//         runner();
//       });
//     },
//   };
// }

// /**
//  * Composes the list of functions passed.
//  *
//  * @param  handlers List of handlers
//  *
//  * @returns function(x) : y
//  */
// export function compose(...handlers: IHandler[]) {
//   return composer('reduceRight', ...handlers);
// }
/**
 * Pipes the list of functions passed
 * @param  handlers List of handlers
 */
function pipe(...handlers) {
    return composer('reduce', ...handlers);
}
/**
 * You may create a pipe | compose based
 * on the array method sent
 *
 * @param method Array method to be called
 * @param handlers list of functions
 *
 * @returns Composed function
 */
function composer(method, ...handlers) {
    return (ctx, ...args) => {
        let hasEnded = false;
        ctx.pipelinePromise = handlers[method]((prev, handler) => __awaiter(this, void 0, void 0, function* () {
            // If the pipeline has ended then
            // do not call further handlers
            if (hasEnded)
                return;
            const result = yield prev;
            // End the pipeline if any handler
            // hasn't returned anything.
            if (result === undefined) {
                hasEnded = true;
                return;
            }
            ctx.data = result;
            return handler(ctx, ...args);
        }), ctx.data);
        return ctx.pipelinePromise;
    };
}

/**
 * Returns all the keys in the Map
 *
 * @param map Map
 *
 * @returns List of keys in the given map
 */
function getMapKeys(map) {
    const keys = [];
    const keysIter = map.keys();
    let key = keysIter.next();
    while (!key.done) {
        keys.push(key.value);
        key = keysIter.next();
    }
    return keys;
}

/**
 * Creates a middleware which will be run
 * only when the pubEvent matches the given
 * pattern
 *
 * @param pattern Subscription event
 * @param middleware Middleware subscribed on the pattern
 */
function patternHandler(pattern, middleware) {
    return (ctx) => {
        if (doesPatternMatch(ctx.pubEvent, pattern)) {
            return middleware(ctx);
        }
        return ctx.data;
    };
}

/**
 * Suffixes internal event part
 * to the given event
 *
 * @param ev Event
 */
const internalEvent = (ev) => `__int_ev__/${ev}`;
/**
 * @hidden
 *
 * Holds all the topics for internal-events
 * such as on init,subscribe,unsubscribe&error
 */
const InternalEvents = {
    ON_INIT: internalEvent('INIT'),
    ON_SUBSCRIBE: internalEvent('ON_SUBSCRIBE'),
    ON_UNSUBSCRIBE: internalEvent('ON_UNSUBSCRIBE'),
    ON_ERROR: internalEvent('ON_ERROR'),
};
/**
 * @hidden
 *
 * This reverse mapping of topic to internal event
 * is used as a predicate to check if the given topic is
 * an internal event or not
 */
const InternalEventsRev = {
    [internalEvent('INIT')]: 'ON_INIT',
    [internalEvent('ON_SUBSCRIBE')]: 'ON_SUBSCRIBE',
    [internalEvent('ON_UNSUBSCRIBE')]: 'ON_UNSUBSCRIBE',
    [internalEvent('ON_ERROR')]: 'ON_ERROR',
};

/**
 * SupervisedEmitter is an event emitter library
 * which supports middlewares, event-tracing, glob subscriptions etc
 *
 * It's main applications can be found in
 * State management, sagas, communication between
 * component irrespective of whereever it is in the DOM tree
 */
class SupervisedEmitter {
    /**
     * Creates a new instance of SupervisedEmitter
     *
     * **Example**
     *
     * Initializing with no middlewares:
     * ```JS
     * import SupervisedEmitter from 'supervised-emitter';
     *
     * const SE = new SupervisedEmitter();
     * ```
     *
     * Initializing with middlewares and options:
     * ```JS
     * const SE = new SupervisedEmitter(
     *   [eventTraceMiddleware],
     *   {debug: true, lfu: {max: 50}}
     * );
     * ```
     *
     * @param middlewares List of middlewares. Remember that all these
     *     middlewares will be piped to form a pipeline. i.e. the output of
     *     each of the middleware is passed in `data`(in the context) to the next
     *     middleware in the pipeline (top-down execution)
     * @param options Options for debugging and LFU
     */
    constructor(middlewares = [({ data }) => data], options = {}) {
        this.logger = new Logger(Logger.LEVEL.DEBUG, false);
        // Pre-compiling regex for efficiency
        this.scopeReg = new RegExp('^__scope_[0-9]+_/(.+)$');
        this.middlewares = pipe(...middlewares);
        this.options = mergeOptions(options, {
            debug: false,
            lfu: {
                max: 100,
                evictCount: 10,
                maxAge: 10 * 60 * 1000,
            },
            publishConcurrency: 100,
            lifeCycleEvents: false,
        });
        this.logger[this.options.debug ? 'enable' : 'disable']();
        // initialize state
        this.state = this.getFreshState();
        this.taskQueue = new TaskQueue(this.publisher.bind(this), { maxRunners: this.options.publishConcurrency });
        this.logger.debug('INITIALIZED');
        this.publishInternalEvents(InternalEvents.ON_INIT, {});
    }
    /**
     * Subscribes to a given event and pipes all the
     * handlers passed for this event.
     *
     * Please note that each handler must pass on the
     * data that must be handled by the next handler, as
     * all these handlers will be piped (compose in reverse direction).
     *
     * Chaining subscriptions is also possible. Please see the
     * example below for more details.
     *
     * For more info on `pipe` visit:
     * https://medium.com/free-code-camp/pipe-and-compose-in-javascript-5b04004ac937
     *
     * **Example**
     *
     * ```JS
     * const subscription = SE.subscribe('foo/bar',
     *   ({data}) => {
     *     console.log(data); //=> 1
     *     return data + 1,
     *   },
     *   ({data}) => {
     *     console.log(data); //=> 2
     *   }
     * ).subscribe('foo/*',
     *   ({data}) => console.log(data) //=> 1
     * );
     *
     * await SE.publish('/foo/bar', 1);
     *
     * subscription.unsubscribe();
     * ```
     *
     * @param event Subscription event
     * @param handlers List of handlers
     *
     * @returns Subscription for chaining more subscriptions or
     *    for unsubscribing from all the subscriptions
     */
    subscribe(event, ...handlers) {
        event = sanitizeEvent(event);
        // Check if this is a new or existing event
        if (!this.state.subscribers.get(event)) {
            // Cache has to be updated if the
            // new pattern event matches with the
            // publish event in cache
            if (isPatternEvent(event)) {
                this.addPatternEventToCache(event);
                this.state.patternEvents.push(event);
            }
            else {
                this.addNormalEventToCache(event);
            }
            this.state.subscribers.set(event, new DLL());
        }
        // We're using dll for maintaining a list
        // of handlers for the ease of splicing
        // handlers during unsubscription
        const eventHandler = this.state.subscribers.get(event).push({
            // Compose all the subscribers passed at once.
            // Users can use this feature if needed, else
            // can choose to go with the classical approach
            // of subscribing as many times as the handlers.
            pipeline: pipe(...handlers),
        });
        this.logger.debug(`SUBSCRIBED => ${event}`);
        // This checks avoid infinite publishing of internal eventss
        if (!this.isInternalEvent(event)) {
            this.publishInternalEvents(InternalEvents.ON_SUBSCRIBE, { subEvent: event });
        }
        const self = this;
        return {
            /**
             * Unsubscribes from this subscription
             */
            unsubscribe() {
                self.unsubscribe(event, eventHandler);
            },
            /**
             * This method allows chaining subscription to
             * multiple events via the same subscription
             */
            subscribe(cEvent, ...cHandlers) {
                const subscription = self.subscribe(cEvent, ...cHandlers);
                const prevSubscription = this;
                return Object.assign(Object.assign({}, subscription), { unsubscribe() {
                        prevSubscription.unsubscribe();
                        subscription.unsubscribe();
                    } });
            },
        };
    }
    /**
     * Similar to [[subscribe]], but it listens only to
     * the first event and unsubscribes itself thereafter.
     *
     * **Example**
     *
     * ```JS
     * let calls = 0;
     * const subscription = SE.subscribeOnce('foo/bar', () => calls++)
     *
     * await SE.publish('/foo/bar', 'test');
     * await SE.publish('/foo/bar', 'test');
     *
     * console.log(calls) //=> 1
     *
     * subscription.unsubscribe();
     * ```
     *
     * @param event Subscription event
     * @param handlers List of handlers
     *
     * @returns Subscription for chaining more subscriptions or
     *    for unsubscribing from all the subscriptions
     */
    subscribeOnce(event, ...handlers) {
        return new Promise((resolve, reject) => {
            const unsubscriber = ({ pipelinePromise, data }) => {
                var _a;
                (_a = pipelinePromise) === null || _a === void 0 ? void 0 : _a.then((sData) => {
                    subscription.unsubscribe();
                    resolve(sData);
                }).catch((err) => {
                    subscription.unsubscribe();
                    reject(err);
                });
                return data;
            };
            const subscription = this.subscribe(event, unsubscriber, ...handlers);
            this.logger.debug(`SUBSCRIBED_ONCE => ${event}`);
        });
    }
    /**
     * Waits untill the required event is received.
     * This is especially useful when writing flows of
     * execution.
     *
     * **Example**
     * In a request life-cycle
     * if some action needs to be take post a response
     * has been received, then it can be written as follow
     * ```JS
     * SE.publish('req/profiles/load')
     * await SE.waitTill('req/profiles/success')
     *
     * SE.publish('req/profiles/sort')
     * ```
     *
     * @param event Subscription event
     *
     * @returns a Promise that resolves to data
     */
    waitTill(event) {
        return this.subscribeOnce(event);
    }
    /**
     * Publishes the given event to all the matching
     * subscribers.
     *
     * NOTE: This is an asynchronous call, so if you want to
     * publish events one after the other, then you will have
     * to `await` on each publish call.
     * Please see the example below for more details.
     *
     * **Example**
     *
     * Simple publish (fire and forget):
     * ```JS
     * SE.publish('foo/bar', 1);
     *
     * SE.publish('foo/bar', 'hello world');
     * ```
     *
     * Publish one after the other (execute all the subscription pipelines before moving to next publish):
     * ```JS
     * await SE.publish('publish/first', 'first');
     *
     * // This will be published only after all the
     * // matching subscription pipelines of the above
     * // publish events have been completed
     * await SE.publish('publish/second', 'second');
     * ```
     *
     * @param pubEvent Event to publish the given data
     * @param data Any data that need to be published
     *
     * @returns Awaitable publish
     */
    publish(pubEvent, data = null) {
        return this.taskQueue.add(pubEvent, data);
    }
    /**
     * Adds scope to a event by prefixing
     * it with a incrementing counter string(__scope_<counter>_/),
     * such that everytime this is called the
     * subscribers can listen only on scoped events.
     * This is especially useful when you don't want
     * other subscribers to listen to this event.
     * Then this behaves more like a camouflage event,
     * which is visible only to scoped subscribers.
     *
     * This is especially useful when multiple
     * instances of the same class is listening and
     * is interested only in events of its own instance.
     *
     * **Example**
     *
     * In React, if you're using the same component in
     * multiple places but your actions(Show popup, make a request etc)
     * are different in each place, then you may achieve it like this:
     * ```JSX
     * /// container.jsx
     * const [{scope}] = useState({scope: SE.getScope()});
     *
     * SE.subscribe(scope('btn/click'), ({data}) => {
     *   // ...
     * });
     *
     * <ChildComponent scope={scope} />
     *
     *
     * /// In ChildComponent.jsx
     * SE.publish(this.props.scope('btn/click'),  data)
     * ```
     *
     * @returns Function(Closure) that can add scope to events
     */
    getScope() {
        const rand = this.state.scopeId++;
        return (event) => `__scope_${rand}_/${event}`;
    }
    /**
     * Strips out the scope part in the given
     * scoped event.
     *
     * i.e, it converts __scope_<number>_/foo/bar => foo/bar
     *
     * This method can be used in your middlewares
     * to unshell the scope part in the topic and run
     * your logics.
     *
     * @param event Scoped event
     *
     * @returns Event without scope part
     */
    unScope(event) {
        const match = event.match(this.scopeReg);
        return match ? match[1] : event;
    }
    /**
     * Returns a fresh / new checkout of
     * state
     */
    getFreshState() {
        return {
            subscribers: new Map(),
            patternEvents: [],
            subEventsCache: new LFUCache(this.options.lfu),
            scopeId: 0,
        };
    }
    /**
     * Goes through every publish event in cache
     * and checks if it matches the given patterns
     * and if match found, then add this pattern
     * to cached events
     *
     * @param patternEvent Pattern event
     */
    addPatternEventToCache(patternEvent) {
        this.state.subEventsCache.forEach(([pubEvent, cachedEvents]) => {
            if (doesPatternMatch(pubEvent, patternEvent)) {
                cachedEvents.set(patternEvent, true);
            }
        });
    }
    /**
     * Checks if the matching pubEvent is present
     * in the cache, if so then it adds this event
     *
     * @param event Normal event (w/o pattern)
     */
    addNormalEventToCache(event) {
        const matchingEvents = this.state.subEventsCache.peek(event);
        if (matchingEvents instanceof Map) {
            matchingEvents.set(event, true);
        }
    }
    /**
     * Unsubscribes the event pipeline from the event
     *
     * @param subEvent Subscribed Event
     * @param eventHandler Handler pipeline
     */
    unsubscribe(subEvent, eventHandler) {
        // This checks avoid infinite publishing of internal events
        if (!this.isInternalEvent(subEvent)) {
            this.publishInternalEvents(InternalEvents.ON_UNSUBSCRIBE, { subEvent });
        }
        const subscribers = this.state.subscribers.get(subEvent);
        // remove the handler from DLL
        if (subscribers) {
            subscribers.remove(eventHandler);
            // If there are no event pipelines
            // for this event, then remove the event from
            // subscribers list (for space optimization).
            if (subscribers.length === 0) {
                this.state.subscribers.delete(subEvent);
            }
        }
        this.logger.debug(`UNSUBSCRIBED => ${subEvent}`);
    }
    /**
     * Returns all the matching subscribed events
     * given publish event (including glob pattern)
     *
     * This is the place where you can optimize the
     * matching logic by caching the matched events.
     *
     * @param pubEvent Publish event
     *
     * @returns map of matching patterns vs this.state
     */
    getSubEvents(pubEvent) {
        let subEvents = this.state.subEventsCache.get(pubEvent);
        if (!subEvents) {
            const matchingEvents = new Map();
            // if normal event subscribers are present
            // then add it as well
            if (this.state.subscribers.get(pubEvent)) {
                matchingEvents.set(pubEvent, true);
            }
            // Check if any pattern matches pubEvent
            this.state.patternEvents.forEach((pattern) => {
                if (doesPatternMatch(pubEvent, pattern)) {
                    matchingEvents.set(pattern, true);
                }
            });
            this.state.subEventsCache.set(pubEvent, matchingEvents);
            subEvents = matchingEvents;
        }
        return subEvents;
    }
    /**
     * Publishes the given data to all the subscribers
     * that match the event.
     *
     * Note: We pass only the second argument (one param)
     * to the subscribers. This is intentionally done
     * because when the eventHandlers are piped (i.e. the output
     * of one is used by the next), incase of which we'll not be
     * able to maintain a standard function signature,
     * if we allow more than one param to be passed to subscribers.
     *
     * @param pubEvent Event to publish the given data
     * @param data Any data that need to be published
     *
     * @returns Promise that resolves after publish pipeline completion
     */
    publisher(pubEvent, data) {
        return __awaiter(this, void 0, void 0, function* () {
            pubEvent = sanitizeEvent(pubEvent);
            const subEvents = this.getSubEvents(pubEvent);
            const subEventsArr = getMapKeys(subEvents);
            const ctx = {
                data,
                pubEvent,
                subEvents: subEventsArr,
            };
            const result = yield this.middlewares(ctx);
            // If the flow was stopped in the middlewares
            // then end it w/o passing the data to any of
            // the subscription pipelines
            if (result === undefined) {
                return;
            }
            ctx.data = result;
            return Promise.all(subEventsArr.map((subEvent) => __awaiter(this, void 0, void 0, function* () {
                // Subscription pipelines
                const subPipelines = this.state.subscribers.get(subEvent);
                // lazy pruning of removed subscription pipelines
                if (!subPipelines) {
                    subEvents.delete(subEvent);
                    return null;
                }
                const pipelinePromises = subPipelines.map(({ pipeline }) => __awaiter(this, void 0, void 0, function* () {
                    try {
                        // use new ctx for every pipeline because
                        // one pipeline must never affect the other
                        // except middleware pipeline, else it gets
                        // difficult to debug
                        const output = yield pipeline(Object.assign({}, ctx));
                        return output;
                    }
                    catch (error) {
                        this.logger.error('Error during publish:', error);
                        this.publishInternalEvents(InternalEvents.ON_ERROR, { error });
                    }
                }));
                return Promise.all(pipelinePromises);
            })));
        });
    }
    publishInternalEvents(pubEvent, data) {
        if (this.isInternalEvent(pubEvent) && this.options.lifeCycleEvents) {
            return this.publish(pubEvent, data);
        }
        return Promise.resolve();
    }
    /**
     * Predicate for determining whether the given
     * event is internal event or not
     *
     * @param event event to be checked for internal event
     *
     * @returns true if the given event is internal event
     */
    isInternalEvent(event) {
        return event in InternalEventsRev;
    }
}
SupervisedEmitter.patternHandler = patternHandler;
SupervisedEmitter.InternalEvents = InternalEvents;

export default SupervisedEmitter;
